@@ -144,14 +144,17 @@ typedef struct
   gpointer user_data;
 
   GtkWidget *window;
+  GtkEventController *window_close_controller;
   gboolean force_closing;
 } AdwDialogPrivate;
 
 static void adw_dialog_buildable_init (GtkBuildableIface *iface);
+static void adw_dialog_shortcut_manager_init (GtkShortcutManagerInterface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (AdwDialog, adw_dialog, GTK_TYPE_WIDGET,
                          G_ADD_PRIVATE (AdwDialog)
-                         G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE, adw_dialog_buildable_init))
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE, adw_dialog_buildable_init)
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_SHORTCUT_MANAGER, adw_dialog_shortcut_manager_init))
 
 static GtkBuildableIface *parent_buildable_iface;
 
@@ -185,6 +188,12 @@ map_tick_cb (AdwDialog *self)
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
 
+  if (priv->force_closing) {
+    priv->tick_cb_id = 0;
+    priv->ticks = 0;
+    return G_SOURCE_REMOVE;
+  }
+
   priv->ticks++;
 
   /* If we're showing a bottom sheet, it has changed after the initial map,
@@ -306,11 +315,14 @@ set_focus (AdwDialog *self,
                                           focus_widget_notify_visible_cb, self);
     g_signal_handlers_disconnect_by_func (priv->focus_widget,
                                           focus_widget_notify_parent_cb, self);
+    g_clear_weak_pointer (&priv->focus_widget);
   }
 
   priv->focus_widget = focus;
 
   if (priv->focus_widget) {
+    g_object_add_weak_pointer (G_OBJECT (priv->focus_widget), (gpointer *) &priv->focus_widget);
+
     g_signal_connect_swapped (priv->focus_widget, "hide",
                               G_CALLBACK (unset_focus_widget), self);
     g_signal_connect_swapped (priv->focus_widget, "notify::visible",
@@ -594,7 +606,6 @@ present_as_window (AdwDialog *self,
 {
   AdwDialogPrivate *priv = adw_dialog_get_instance_private (self);
   GtkWidget *titlebar;
-  GtkEventController *shortcut_controller;
   GtkShortcut *shortcut;
 
   if (priv->window) {
@@ -613,9 +624,10 @@ present_as_window (AdwDialog *self,
   shortcut = gtk_shortcut_new (gtk_keyval_trigger_new (GDK_KEY_Escape, 0),
                                gtk_callback_action_new ((GtkShortcutFunc) maybe_close_cb, self, NULL));
 
-  shortcut_controller = gtk_shortcut_controller_new ();
-  gtk_shortcut_controller_add_shortcut (GTK_SHORTCUT_CONTROLLER (shortcut_controller), shortcut);
-  gtk_widget_add_controller (priv->window, shortcut_controller);
+  priv->window_close_controller = gtk_shortcut_controller_new ();
+  gtk_shortcut_controller_set_scope (GTK_SHORTCUT_CONTROLLER (priv->window_close_controller), GTK_SHORTCUT_SCOPE_MANAGED);
+  gtk_shortcut_controller_add_shortcut (GTK_SHORTCUT_CONTROLLER (priv->window_close_controller), shortcut);
+  gtk_widget_add_controller (GTK_WIDGET (self), priv->window_close_controller);
 
   if (parent) {
     GtkRoot *root = gtk_widget_get_root (parent);
@@ -767,6 +779,23 @@ ensure_focus (AdwDialog *self)
   return TRUE;
 }
 
+static gboolean
+toggle_adaptive_preview_cb (AdwDialog *self)
+{
+  GtkRoot *root;
+  gboolean open;
+
+  if (!adw_get_inspector_keybinding_enabled ())
+    return GDK_EVENT_PROPAGATE;
+
+  root = gtk_widget_get_root (GTK_WIDGET (self));
+
+  g_object_get (root, "adaptive-preview", &open, NULL);
+  g_object_set (root, "adaptive-preview", !open, NULL);
+
+  return GDK_EVENT_STOP;
+}
+
 static void
 adw_dialog_root (GtkWidget *widget)
 {
@@ -861,9 +890,11 @@ adw_dialog_dispose (GObject *object)
                                           focus_widget_notify_visible_cb, self);
     g_signal_handlers_disconnect_by_func (priv->focus_widget,
                                           focus_widget_notify_parent_cb, self);
-    priv->focus_widget = NULL;
+    g_clear_weak_pointer (&priv->focus_widget);
   }
 
+  g_clear_weak_pointer (&priv->last_focus);
+
   if (priv->default_widget) {
     g_signal_handlers_disconnect_by_func (priv->default_widget,
                                           unset_default_widget, self);
@@ -881,7 +912,8 @@ adw_dialog_dispose (GObject *object)
     priv->floating_sheet = NULL;
     priv->child_breakpoint_bin = NULL;
     priv->child = NULL;
-  } else if (priv->window) {
+  } else if (priv->child_breakpoint_bin &&
+             gtk_widget_get_parent (priv->child_breakpoint_bin) == GTK_WIDGET (self)) {
     /* It's an window-backed dialog */
     g_clear_pointer (&priv->child_breakpoint_bin, gtk_widget_unparent);
     priv->child = NULL;
@@ -993,6 +1025,16 @@ adw_dialog_set_property (GObject      *object,
   }
 }
 
+static void
+adw_dialog_real_closed (AdwDialog *dialog)
+{
+}
+
+static void
+adw_dialog_real_close_attempt (AdwDialog *dialog)
+{
+}
+
 static void
 adw_dialog_class_init (AdwDialogClass *klass)
 {
@@ -1012,6 +1054,9 @@ adw_dialog_class_init (AdwDialogClass *klass)
   widget_class->contains = adw_widget_contains_passthrough;
   widget_class->compute_expand = adw_widget_compute_expand;
 
+  klass->closed = adw_dialog_real_closed;
+  klass->close_attempt = adw_dialog_real_close_attempt;
+
   /**
    * AdwDialog:child:
    *
@@ -1239,6 +1284,8 @@ adw_dialog_class_init (AdwDialogClass *klass)
                                 (GtkShortcutFunc) open_inspector_cb, "b", FALSE);
   gtk_widget_class_add_binding (widget_class, GDK_KEY_D, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
                                 (GtkShortcutFunc) open_inspector_cb, "b", TRUE);
+  gtk_widget_class_add_binding (widget_class, GDK_KEY_M, GDK_CONTROL_MASK | GDK_SHIFT_MASK,
+                                (GtkShortcutFunc) toggle_adaptive_preview_cb, NULL);
 
   gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
   gtk_widget_class_set_css_name (widget_class, "dialog");
@@ -1258,6 +1305,8 @@ adw_dialog_init (AdwDialog *self)
   priv->follows_content_size = FALSE;
   priv->presentation_mode = ADW_DIALOG_AUTO;
 
+  gtk_widget_set_limit_events (GTK_WIDGET (self), TRUE);
+
   priv->child_breakpoint_bin = adw_breakpoint_bin_new ();
   gtk_widget_set_overflow (priv->child_breakpoint_bin, GTK_OVERFLOW_VISIBLE);
   adw_breakpoint_bin_set_warning_widget (ADW_BREAKPOINT_BIN (priv->child_breakpoint_bin),
@@ -1293,6 +1342,11 @@ adw_dialog_buildable_init (GtkBuildableIface *iface)
   iface->add_child = adw_dialog_buildable_add_child;
 }
 
+static void
+adw_dialog_shortcut_manager_init (GtkShortcutManagerInterface *iface)
+{
+}
+
 /**
  * adw_dialog_new:
  *
@@ -1348,14 +1402,14 @@ adw_dialog_set_child (AdwDialog *self,
   g_return_if_fail (ADW_IS_DIALOG (self));
   g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
 
-  if (child)
-    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
-
   priv = adw_dialog_get_instance_private (self);
 
   if (priv->child == child)
     return;
 
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
+
   priv->child = child;
 
   adw_breakpoint_bin_set_child (ADW_BREAKPOINT_BIN (priv->child_breakpoint_bin), child);
@@ -1876,14 +1930,29 @@ adw_dialog_close (AdwDialog *self)
 
   priv = adw_dialog_get_instance_private (self);
 
+  if (!gtk_widget_get_parent (GTK_WIDGET (self))) {
+    g_critical ("Trying to close %s %p that's not presented", G_OBJECT_TYPE_NAME (self), self);
+    return FALSE;
+  }
+
   if (!priv->can_close) {
     g_signal_emit (self, signals[SIGNAL_CLOSE_ATTEMPT], 0);
     return FALSE;
   }
 
   if (priv->window) {
-    gtk_window_close (GTK_WINDOW (priv->window));
+    GtkWidget *window = priv->window;
     priv->window = NULL;
+
+    gtk_widget_remove_controller (GTK_WIDGET (self), priv->window_close_controller);
+    priv->window_close_controller = NULL;
+
+    if (priv->closing_callback)
+      priv->closing_callback (self, priv->user_data);
+
+    g_signal_emit (self, signals[SIGNAL_CLOSED], 0);
+
+    gtk_window_close (GTK_WINDOW (window));
   } else {
     adw_dialog_force_close (self);
   }
